import logging
import unittest.mock as mock
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pytest
from astropy.table import Table
from torchvision.transforms.v2 import CenterCrop, Lambda

from hyrax.data_sets.hsc_data_set import FitsImageDataSet, HSCDataSet

test_dir = Path(__file__).parent / "test_data" / "dataloader"


@pytest.fixture(scope="module")
def hsc_test_mode():
    """Turn on unit test mode for HSC and FitsImage datasets. This allows them to run without metadata"""
    HSCDataSet._called_from_test = True
    FitsImageDataSet._called_from_test = True


class FakeFitsFS:
    """
    Mocks the only operations on a directory full of fits files that
    the dataloader should care about for initialization:

    1) Globbing fits files with Path.glob

    2) Reading the shape of the first table out of the fits header using
    astropy.io.fits.open

    If we end up doing anything more sophisticated this will need to be changed
    We are planning to scan 1M - 10M fits files so its unlikely we want to do
    more filesystem operations without a really good reason.
    """

    def __init__(self, test_files: dict, filter_catalog: Optional[dict] = None):
        self.patchers: list[mock._patch[Any]] = []

        self.test_files = test_files

        mock_paths = [Path(x) for x in list(test_files.keys())]
        target = "hyrax.data_sets.hsc_data_set.Path.iterdir"
        self.patchers.append(mock.patch(target, return_value=mock_paths))

        mock_fits_open = mock.Mock(side_effect=self._open_file)
        self.patchers.append(mock.patch("astropy.io.fits.open", mock_fits_open))

        if filter_catalog is not None:
            mock_read_filter_catalog = mock.patch(
                "hyrax.data_sets.hsc_data_set.HSCDataSet._read_filter_catalog",
                lambda x, y: Table(
                    {"object_id": ["notreal"], "filter": ["notreal"], "filename": ["notreal"]}
                ),
            )
            self.patchers.append(mock_read_filter_catalog)
            mock_parse_filter_catalog = mock.patch(
                "hyrax.data_sets.hsc_data_set.FitsImageDataSet._parse_filter_catalog",
                lambda x, y: filter_catalog,
            )
            self.patchers.append(mock_parse_filter_catalog)

    def _open_file(self, filename: Path, **kwargs) -> mock.Mock:
        shape = self.test_files[filename.name]
        mock_open_ctx = mock.Mock()
        mock_open_ctx.__enter__ = mock.Mock(return_value=["", np.zeros(shape)])
        mock_open_ctx.__exit__ = mock.Mock()
        return mock_open_ctx

    def __enter__(self):
        for patcher in self.patchers:
            patcher.start()

    def __exit__(self, *exc):
        for patcher in self.patchers:
            patcher.stop()


def mkconfig(
    crop_to=False,
    filters=False,
    filter_catalog=False,
    use_cache=False,
    transform="tanh",
):
    """Makes a configuration that points at nonexistent path so HSCDataSet.__init__ will create an object,
    and our FakeFitsFS shim can be called.
    """
    return {
        "general": {"data_dir": "thispathdoesnotexist"},
        "data_set": {
            "crop_to": crop_to,
            "filters": filters,
            "filter_catalog": filter_catalog,
            "use_cache": use_cache,
            "transform": transform,
            "preload_cache": False,  # Don't run the preloading in unit tests, because it needs real data.
            "object_id_column_name": False,
            "filter_column_name": False,
            "filename_column_name": False,
        },
    }


def generate_files(
    num_objects=10, num_filters=5, shape=(100, 100), offset=0, infill_str="all_filters"
) -> dict:
    """Generates a dictionary to pass in to FakeFitsFS.

    This generates a dict from filename->shape tuple for a set of uniform fake fits files
    corresponding to the naming convention used by the HSC data loader.

    Completely uniform data sets with all filters for all files with every file the same
    size are generated by this function. Sequential object_ids are used starting from 0
    by default.

    Parameters
    ----------
    num_objects : int, optional
        How many objects are represented in the data set, by default 10
    num_filters : int, optional
        How many filters does each object have, by default 5. If you provide a number greater than 5,
        only 5 filters will be output.
    shape : tuple, optional
        What are the dimensions of the image in each fits file, by default (100,100)
    offset : int, optional
        What is the first object_id to start with, by default 0
    infill_str: str, optional
        What to put in the fake filename in between the object ID and filter name. By default "all_filters"

    Returns
    -------
    dict
        Dictionary from filename -> shape appropriate for FakeFitsFS
    """
    filters = ["HSC-G", "HSC-R", "HSC-I", "HSC-Z", "HSC-Y"][:num_filters]
    test_files = {}
    for object_id in range(offset, num_objects + offset):
        for filter in filters:
            test_files[f"{object_id:017d}_{infill_str}_{filter}.fits"] = shape

    return test_files


def generate_filter_catalog(test_files: dict, config: dict) -> dict:
    """Generates a filter catalog dict for use with FakeFitsFS from a filesystem dictionary
    created by generate_files.

    This allows tests to alter the parsed filter_catalog, and interrogate what decisions HSCDataSet makes
    when a manifest or filter_catalog file contains corrupt information.

    Caveats:
    Always Run this prior to creating any mocks, or using a FakeFitsFS

    This function calls the HSCDataSet parsing code to actually assemble the files array, which means:
    1) This test setup is fairly tightly coupled to the internals of HSCDataSet
    2) When writing tests using this, ensure any functionality you depend on from the slow file-scan
       initialization codepath is tested indepently.

    Parameters
    ----------
    test_files : dict
        Test files dictionary created with generate_files.
    config : dict
        The config you will use to initialize the test HSCDataSet object in your test.
        Create this with mkconfig()

    Returns
    -------
    dict
        Dictionary from ObjectID -> (Dictionary from Filter -> Filename)
    """
    # Ensure the filter_catalog codepath won't be triggered
    config["data_set"]["filter_catalog"] = False

    # Use our initialization code to create a parsed files object
    with FakeFitsFS(test_files):
        return HSCDataSet(config).files


def test_load(caplog, hsc_test_mode):
    """Test to ensure loading a perfectly regular set of files works"""
    caplog.set_level(logging.WARNING)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))
    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        # 10 objects should load
        assert len(a) == 10

        # The number of filters, and image dimensions should be correct and square
        assert a.shape() == (5, 262, 262)

        # No warnings should be printed
        assert caplog.text == ""


def test_load_duplicate(caplog, hsc_test_mode):
    """Test to ensure duplicate fits files that reference the same object id and filter create the
    appropriate error messages.
    """
    caplog.set_level(logging.ERROR)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))
    duplicate_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263), infill_str="duplicate")
    test_files.update(duplicate_files)
    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        # Only 10 objects should load
        assert len(a) == 10

        # The number of filters, and image dimensions should be correct and square
        assert a.shape() == (5, 262, 262)

        # We should get duplicate object errors
        assert "Duplicate object ID" in caplog.text

        # We should get errors that include the duplicate filenames
        assert "_duplicate_" in caplog.text

        # The duplicate files should not be in the data set
        for filepath in a._all_files():
            assert "_duplicate_" not in str(filepath)


def test_prune_warn_1_percent(caplog, hsc_test_mode):
    """Test to ensure when >1% of loaded objects are missing a filter, that is a warning
    and that the resulting dataset drops the objects that are missing filters
    """
    caplog.set_level(logging.WARNING)

    # Generate two files which
    test_files = generate_files(num_objects=98, num_filters=2, shape=(100, 100))
    # Object 101 is missing the HSC-G filter, we only provide the R filter
    test_files["00000000000000101_missing_g_HSC-R.fits"] = (100, 100)

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        # We should have the correct number of objects
        assert len(a) == 98

        # Object 2 should not be loaded
        assert "00000000000000101" not in a

        # We should Error log because greater than 5% of the objects were pruned
        assert "Greater than 1% of objects in the data directory were pruned." in caplog.text

        # We should warn that we dropped an object explicitly
        assert "Dropping object" in caplog.text


def test_prune_error_5_percent(caplog, hsc_test_mode):
    """Test to ensure when >5% of loaded objects are missing a filter, that is an error
    and that the resulting dataset drops the objects that are missing filters
    """
    caplog.set_level(logging.ERROR)

    # Generate two files which
    test_files = generate_files(num_objects=18, num_filters=2, shape=(100, 100))
    # Object 20 is missing the HSC-G filter, we only provide the R filter
    test_files["00000000000000020_missing_g_HSC-R.fits"] = (100, 100)

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        # We should have two objects, having dropped one.
        assert len(a) == 18

        # Object 20 should not be loaded
        assert "00000000000000020" not in a

        # We should Error log because greater than 5% of the objects were pruned
        assert "Greater than 5% of objects in the data directory were pruned." in caplog.text


def test_crop(caplog, hsc_test_mode):
    """Test to ensure that in the presence of heterogenous sizes within 1px of a central size
    We load all images and crop to the smallest dimenensions without any logs
    """
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))
    # Add some images with dimensions 1 px larger
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(101, 100), offset=10))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 101), offset=20))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(101, 101), offset=30))
    # Add some images with dimensions 1 px smaller
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(99, 100), offset=40))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 99), offset=50))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(99, 99), offset=60))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        assert len(a) == 70
        assert a.shape() == (5, 99, 99)

        # No warnings should be printed since we're within 1px of the mean size
        assert caplog.text == ""


def test_crop_warn_2px_larger(caplog, hsc_test_mode):
    """Test to ensure that in the presence of heterogenous sizes within 2px of a central size
    We load all images and crop to the smallest dimenensions and warn the user of the issue
    """
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))
    # Add some images with dimensions 2 px larger
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(102, 100), offset=10))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 102), offset=20))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(102, 102), offset=30))
    # Add some images with dimensions 1 px smaller
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(99, 100), offset=40))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 99), offset=50))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(99, 99), offset=60))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        assert len(a) == 70
        assert a.shape() == (5, 99, 99)

        # We should warn that images differ
        assert "Some images differ" in caplog.text


def test_crop_warn_2px_smaller(caplog, hsc_test_mode):
    """Test to ensure that in the presence of heterogenous sizes within 2px of a central size
    We load all images and crop to the smallest dimenensions and warn the user of the issue
    """
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))
    # Add some images with dimensions 1 px larger
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(101, 100), offset=10))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 101), offset=20))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(101, 101), offset=30))
    # Add some images with dimensions 2 px smaller
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(98, 100), offset=40))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 98), offset=50))
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(98, 98), offset=60))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig())

        assert len(a) == 70
        assert a.shape() == (5, 98, 98)

        # We should warn that images differ
        assert "Some images differ" in caplog.text


def test_prune_size(caplog, hsc_test_mode):
    """Test to ensure images that are too small will be pruned from the data set when a custom size is
    passed."""
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))
    # Add some images with dimensions 1 px larger
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(101, 101), offset=20))
    # Add some images with dimensions 2 px smaller
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(98, 98), offset=30))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(crop_to=(99, 99)))

        assert len(a) == 20
        assert a.shape() == (5, 99, 99)

        # We should warn that we are dropping objects and the reason
        assert "Dropping object" in caplog.text
        assert "too small" in caplog.text


def test_prune_filter_size_mismatch(caplog, hsc_test_mode):
    """Test to ensure images with different sizes per filter will be dropped"""
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))
    test_files["00000000000000000_all_filters_HSC-R.fits"] = (99, 99)

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(crop_to=(99, 99)))

        assert len(a) == 9
        assert a.shape() == (5, 99, 99)

        # We should warn that we are dropping objects and the reason
        assert "Dropping object" in caplog.text
        assert "first filter" in caplog.text


def test_prune_bad_filename(caplog, hsc_test_mode):
    """Test to ensure images with filenames set wrong will be dropped"""
    caplog.set_level(logging.WARNING)
    test_files = {}
    test_files.update(generate_files(num_objects=10, num_filters=5, shape=(100, 100), offset=0))

    config = mkconfig(crop_to=(99, 99), filter_catalog="notarealfile.fits")

    # Create a filter catalog with wrong file information.
    filter_catalog = generate_filter_catalog(test_files, config)
    filters = list(filter_catalog["00000000000000000"].keys())
    filter_catalog["00000000000000000"][filters[0]] = filter_catalog["00000000000000001"][filters[0]]

    with FakeFitsFS(test_files, filter_catalog):
        # Initialize HSCDataset exercising the filter_catalog provided initialization pathway
        a = HSCDataSet(config)

        # Verify that the broken object has been dropped
        assert len(a) == 9

        # Verify the shape is correct.
        assert a.shape() == (5, 99, 99)

        # We should warn that we are dropping objects and the correct reason
        assert "Dropping object" in caplog.text
        assert "manifest is likely corrupt" in caplog.text


def test_partial_filter(caplog, hsc_test_mode):
    """Test to ensure when we only load some of the filters, only those filters end up in the dataset"""
    caplog.set_level(logging.WARNING)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))
    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(filters=["HSC-G", "HSC-R"]))

        # 10 objects should load
        assert len(a) == 10

        # The number of filters, and image dimensions should be correct and square
        assert a.shape() == (2, 262, 262)

        # No warnings should be printed
        assert caplog.text == ""


def test_partial_filter_prune_warn_1_percent(caplog, hsc_test_mode):
    """Test to ensure when a the user supplies a filter list and >1% of loaded objects are
    missing a filter, that is a warning and that the resulting dataset drops the objects that
    are missing filters.
    """
    caplog.set_level(logging.WARNING)

    # Generate two files which
    test_files = generate_files(num_objects=98, num_filters=3, shape=(100, 100))
    # Object 101 is missing the HSC-G and HSC-I filters, we only provide the R filter
    test_files["00000000000000101_missing_g_HSC-R.fits"] = (100, 100)

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(filters=["HSC-R", "HSC-I"]))

        # We should have the correct number of objects
        assert len(a) == 98

        # Object 101 should not be loaded
        assert "00000000000000101" not in a

        # We should Error log because greater than 5% of the objects were pruned
        assert "Greater than 1% of objects in the data directory were pruned." in caplog.text

        # We should warn that we dropped an object explicitly
        assert "Dropping object" in caplog.text


def test_valid_transform_string(caplog, hsc_test_mode):
    """Test to ensure that a valid string passed to transform
    will map to a numpy function"""

    caplog.set_level(logging.ERROR)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(transform="arcsinh"))

        # transform always has CenterCrop in the beginning followed by the user
        # defined transform
        lambda_transform = [t for t in a.transform.transforms if isinstance(t, Lambda)][0]
        assert lambda_transform.lambd == np.arcsinh

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(transform="tanh"))

        # transform always has CenterCrop in the beginning followed by the user
        # defined transform
        lambda_transform = [t for t in a.transform.transforms if isinstance(t, Lambda)][0]
        assert lambda_transform.lambd == np.tanh


def test_invalid_transform_string(caplog, hsc_test_mode):
    """Test to ensure that an invalid string passed to transform will raise an error"""

    caplog.set_level(logging.ERROR)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))

    with FakeFitsFS(test_files):
        with pytest.raises(RuntimeError):
            HSCDataSet(mkconfig(transform="invalid_function"))


def test_false_transform(caplog, hsc_test_mode):
    """Test to ensure that false passed to transform behaves as expected"""

    caplog.set_level(logging.ERROR)
    test_files = generate_files(num_objects=10, num_filters=5, shape=(262, 263))

    with FakeFitsFS(test_files):
        a = HSCDataSet(mkconfig(transform=False))

        # When transform is False; only a CenterCrop should be applied
        # automatically with a size conforming to test_files above
        expected_transform = CenterCrop(size=(np.int64(262), np.int64(262)))
        actual_transform = a.transform
        assert isinstance(actual_transform, CenterCrop)
        assert actual_transform.size == expected_transform.size



from .config_utils import log_runtime_config
from .hyrax import Hyrax
from .plugin_utils import get_or_load_class, import_module_from_string, update_registry

__all__ = [
    "log_runtime_config",
    "get_or_load_class",
    "import_module_from_string",
    "update_registry",
    "Hyrax",
]



# ruff: noqa: D101, D102

# This is a more felxible version of hsc_dcae.py that should
# work with a variety of image sizes.

import torch
import torch.nn as nn
import torch.nn.functional as f

# extra long import here to address a circular import issue
from hyrax.models.model_registry import hyrax_model


class ArcsinhActivation(nn.Module):
    """Helper module for HSCDAEv2 to use the arcsinh function"""

    def forward(self, x):
        return torch.arcsinh(x)


@hyrax_model
class HSCDCAEv2(nn.Module):
    """
    This autoencoder is designed to work with datasets that are prepared with Hyrax's HSC Data Set class.
    """

    def __init__(self, config, shape):
        super().__init__()

        # Store input shape for dynamic sizing
        self.input_shape = shape

        # Extract number of channels from input shape (assuming NCHW format)
        in_channels = shape[1] if len(shape) > 3 else shape[0]

        # Encoder
        self.encoder1 = nn.Conv2d(in_channels, 32, kernel_size=3, stride=1, padding=1)
        self.encoder2 = nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1)
        self.encoder3 = nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1)
        self.encoder4 = nn.Conv2d(128, 256, kernel_size=3, stride=1, padding=1)

        self.pool = nn.MaxPool2d(2, 2)

        # Decoder - using normal Conv2d with upsampling instead of ConvTranspose2d
        # This approach is more flexible for different image sizes
        self.decoder4 = nn.Conv2d(256, 128, kernel_size=3, stride=1, padding=1)
        self.decoder3 = nn.Conv2d(128, 64, kernel_size=3, stride=1, padding=1)
        self.decoder2 = nn.Conv2d(64, 32, kernel_size=3, stride=1, padding=1)
        self.decoder1 = nn.Conv2d(32, in_channels, kernel_size=3, stride=1, padding=1)

        self.activation = nn.ReLU()

        # Configure final activation
        final_layer = config["model"].get("HSCDCAE_final_layer", "identity")
        if final_layer == "sigmoid":
            self.final_activation = nn.Sigmoid()
        elif final_layer == "tanh":
            self.final_activation = nn.Tanh()
        elif final_layer == "arcsinh":
            self.final_activation = ArcsinhActivation()
        else:
            self.final_activation = nn.Identity()

        self.config = config

    def encode(self, x):
        # Encoder with skip connections
        x1 = self.activation(self.encoder1(x))
        p1 = self.pool(x1)

        x2 = self.activation(self.encoder2(p1))
        p2 = self.pool(x2)

        x3 = self.activation(self.encoder3(p2))
        p3 = self.pool(x3)

        x4 = self.activation(self.encoder4(p3))

        return x4, [x3, x2, x1]

    def decode(self, x, skip_connections):
        # Decoder with skip connections and dynamic upsampling
        x = f.interpolate(x, size=skip_connections[0].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder4(x) + skip_connections[0])

        x = f.interpolate(x, size=skip_connections[1].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder3(x) + skip_connections[1])

        x = f.interpolate(x, size=skip_connections[2].shape[2:], mode="bilinear", align_corners=False)
        x = self.activation(self.decoder2(x) + skip_connections[2])

        # Final interpolation to input size
        if hasattr(self, "original_size"):
            x = f.interpolate(x, size=self.original_size, mode="bilinear", align_corners=False)

        x = self.final_activation(self.decoder1(x))

        return x

    def forward(self, x):
        # Dropping labels if present
        x = x[0] if isinstance(x, tuple) else x

        # Store original spatial dimensions for decoding
        self.original_size = x.shape[2:]

        # Encode
        encoded, skip_connections = self.encode(x)

        return encoded

    def train_step(self, batch):
        """This function contains the logic for a single training step.

        Parameters
        ----------
        batch : tuple
            A tuple containing the two values the loss function

        Returns
        -------
        Current loss value : dict
            Dictionary containing the loss value for the current batch.
        """
        # Dropping labels if present
        data = batch[0] if isinstance(batch, tuple) else batch
        self.optimizer.zero_grad()

        # Store original spatial dimensions for decoding
        self.original_size = data.shape[2:]

        # Encode
        encoded, skip_connections = self.encode(data)

        # Decode
        decoded = self.decode(encoded, skip_connections)

        # Compute loss
        loss = self.criterion(decoded, data)
        loss.backward()
        self.optimizer.step()

        return {"loss": loss.item()}



import numpy as np
import pytest
from astropy.table import Table
from torch import from_numpy
from torch.utils.data import Dataset, IterableDataset

import hyrax
from hyrax.data_sets import HyraxDataset


class RandomDataset(HyraxDataset, Dataset):
    """Dataset yielding pairs of random numbers. Requires a seed to emulate
    static data on the filesystem between instantiations"""

    def __init__(self, config):
        size = config["data_set"]["size"]

        dim_1_length = 2
        if "dimension_1_length" in config["data_set"]:
            dim_1_length = config["data_set"]["dimension_1_length"]

        dim_2_length = 0
        if "dimension_2_length" in config["data_set"]:
            dim_2_length = config["data_set"]["dimension_2_length"]

        seed = config["data_set"]["seed"]
        rng = np.random.default_rng(seed)

        print(f"Initialized dataset with dim 1: {dim_1_length}, dim 2: {dim_2_length}")

        if dim_2_length > 0:
            self.data = rng.random((size, dim_1_length, dim_2_length), np.float32)
        else:
            self.data = rng.random((size, dim_1_length), np.float32)

        # Start our IDs at a random integer between 0 and 100
        id_start = rng.integers(100)
        self.id_list = list(range(id_start, id_start + size))

        metadata_table = Table({"object_id": np.array(list(self.ids()))})

        super().__init__(config, metadata_table)

    def __getitem__(self, idx):
        return from_numpy(self.data[idx])

    def __len__(self):
        return len(self.data)

    def ids(self):
        """Yield IDs for the dataset"""
        for id_item in self.id_list:
            yield str(id_item)


class RandomIterableDataset(RandomDataset, IterableDataset):
    """Iterable version of RandomDataset"""

    def __iter__(self):
        for item in self.data:
            yield from_numpy(item)


@pytest.fixture(scope="function", params=["HyraxRandomDataset", "HyraxRandomIterableDataset"])
def loopback_hyrax(tmp_path_factory, request):
    """This generates a loopback hyrax instance
    which is configured to use the loopback model
    and a simple dataset yielding random numbers
    """
    results_dir = tmp_path_factory.mktemp(f"loopback_hyrax_{request.param}")

    h = hyrax.Hyrax()
    h.config["model"]["name"] = "HyraxLoopback"
    h.config["train"]["epochs"] = 1
    h.config["data_loader"]["batch_size"] = 5
    h.config["general"]["results_dir"] = str(results_dir)

    h.config["general"]["dev_mode"] = True
    h.config["data_set"]["name"] = request.param
    h.config["data_set.random_dataset"]["size"] = 20
    h.config["data_set.random_dataset"]["seed"] = 0
    h.config["data_set.random_dataset"]["shape"] = [2, 3]

    h.config["data_set"]["validate_size"] = 0.2
    h.config["data_set"]["test_size"] = 0.2
    h.config["data_set"]["train_size"] = 0.6

    weights_file = results_dir / "fakeweights"
    with open(weights_file, "a"):
        pass
    h.config["infer"]["model_weights_file"] = str(weights_file)

    dataset = h.prepare()
    return h, dataset


@pytest.fixture(scope="function")
def loopback_inferred_hyrax(loopback_hyrax):
    """This generates a loopback hyrax instance which is configured to use the
    loopback model and a simple dataset yielding random numbers. It includes a call
    to hyrax.infer which will produce the output consumed by vdb_index or umap."""

    h, dataset = loopback_hyrax
    inference_results = h.infer()

    return h, dataset, inference_results



import logging

from hyrax.pytorch_ignite import setup_dataset

logger = logging.getLogger(__name__)


def run(config):
    """Rebuild a broken download manifest

    Parameters
    ----------
    config : dict
        The parsed config file as a nested
        dict
    """
    from .data_sets.hsc_data_set import HSCDataSet

    config["rebuild_manifest"] = True

    data_set = setup_dataset(config)

    if not isinstance(data_set, HSCDataSet):
        msg = "Invalid to run rebuild manafest except on an HSCDataSet."
        raise RuntimeError(msg)

    logger.info("Starting rebuild of manifest")

    data_set._rebuild_manifest(config)

    logger.info("Finished Rebuild Manifest")



from typing import Union

from hyrax.vector_dbs.vector_db_interface import VectorDB


def vector_db_factory(config: dict, context: dict) -> Union[VectorDB, None]:
    """Factory method to create a database object"""

    # if the vector_db name is `False`, return None
    if not config["vector_db"]["name"]:
        return None

    vector_db_name = config["vector_db"]["name"]

    if vector_db_name == "chromadb":
        from hyrax.vector_dbs.chromadb_impl import ChromaDB

        return ChromaDB(config, context)
    elif vector_db_name == "qdrant":
        from hyrax.vector_dbs.qdrantdb_impl import QdrantDB

        return QdrantDB(config, context)
    else:
        return None



# Remove import sorting, these are imported in the order written so that
# autoapi docs are generated with ordering controlled below.
# ruff: noqa: I001
from .hsc_autoencoder import HSCAutoencoder
from .hsc_dcae import HSCDCAE
from .hsc_dcae_v2 import HSCDCAEv2
from .hyrax_autoencoder import HyraxAutoencoder
from .hyrax_cnn import HyraxCNN
from .hyrax_loopback import HyraxLoopback
from .model_registry import hyrax_model

__all__ = [
    "hyrax_model",
    "HyraxAutoencoder",
    "HyraxCNN",
    "HyraxLoopback",
    "HSCAutoencoder",
    "HSCDCAE",
    "HSCDCAEv2",
]



import logging
from pathlib import Path

import onnx
import onnxruntime
from numpy import allclose

logger = logging.getLogger(__name__)


def export_to_onnx(model, sample, config, ctx):
    """Dispatching function to convert a ML framework model into an ONNX model.

    Parameters
    ----------
    model : ML framework model
        The model that was just trained using the ML framework. i.e. PyTorch
    sample : Tensor
        A single sample from the training data loader. This is used to check the
        output of the ONNX model against the output of the PyTorch model.
    config : dict
        The parsed config file as a nested dict
    ctx : dict
        A context dictionary containing info needed for the conversion to ONNX.
    """

    # build the output ONNX file path
    model_filename = Path(config["train"]["weights_filepath"]).stem
    onnx_opset_version = config["onnx"]["opset_version"]
    onnx_model_filename = f"{model_filename}_opset_{onnx_opset_version}.onnx"
    onnx_output_filepath = ctx["results_dir"] / onnx_model_filename

    # use the "ml_framework" context value to determine how to convert to ONNX.
    sample_out = None
    if ctx["ml_framework"] == "pytorch":
        sample, sample_out = _export_pytorch_to_onnx(model, sample, onnx_output_filepath, onnx_opset_version)
    else:
        logger.warning("No ONNX export implementation for the given ML framework.")
        return

    # check the ONNX model for correctness
    try:
        onnx_model = onnx.load(onnx_output_filepath)
        onnx.checker.check_model(onnx_model)
    except:  # noqa E722
        logger.error(f"Failed to create ONNX model. {ctx['ml_framework']} implementation has been saved.")

    # Check the ONNX model against the PyTorch model. Note that `sample` was
    # converted to numpy array when the model was converted to ONNX
    ort_session = onnxruntime.InferenceSession(onnx_output_filepath)
    ort_inputs = {ort_session.get_inputs()[0].name: sample}
    ort_outs = ort_session.run(None, ort_inputs)

    # verify ONNX model inference produces results close to the the original model
    if not allclose(sample_out, ort_outs[0], rtol=1e-03, atol=1e-05):
        logger.warning("The outputs from the PyTorch model and the ONNX model are not close.")

    logger.info(f"Exported model to ONNX format: {onnx_output_filepath}")


def _export_pytorch_to_onnx(model, sample, output_filepath, opset_version):
    """Specific implementation to convert PyTorch model to ONNX format. This
    function will also:
    -  Run `sample` through the model before converting the model to ONNX
    -  Convert `sample` to a numpy array
    """

    # deferred import to reduce start up time
    from torch.onnx import export

    # set model in eval mode and move it to the CPU to prep for export to ONNX.
    model.train(False)
    model.to("cpu")

    # run a single sample through the model. We'll check this against the output
    # from the ONNX version to make sure it's the same, i.e. `np.assert_allclose`.
    sample_out = model(sample)

    # export the model to ONNX format
    export(
        model,
        sample,
        output_filepath,
        opset_version=opset_version,
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
    )

    # return the input sample as a numpy array and the output of the sample run
    # through the model as numpy array
    return sample.numpy(), sample_out.detach().numpy()



import os
import uuid
from typing import Union

import numpy as np
from qdrant_client import QdrantClient, models

from hyrax.vector_dbs.vector_db_interface import VectorDB


class QdrantDB(VectorDB):
    """Implementation of the VectorDB interface using Qdrant as the backend."""

    def __init__(self, config, context):
        super().__init__(config, context)
        self.client = None
        self.collection_size = 0

    def _convert_id_to_uuid(self, id: Union[str, int]) -> str:
        """Convert an id to a UUID string using the OID namespace."""
        return uuid.uuid5(uuid.NAMESPACE_OID, str(id)).urn

    def connect(self):
        """Connect to the Qdrant database and return an instance of the client."""
        # Results_dir is the directory where the Qdrant database is stored.
        results_dir = self.context["results_dir"]
        self.client = QdrantClient(path=results_dir)
        return self.client

    def create(self):
        """Create a new Qdrant database"""
        if self.client is None:
            self.connect()

        # We'll get the number of collection that are in the db, but for now
        # we follow the advice of the documentation, and restrict the database
        # to a single collection.
        # https://qdrant.tech/documentation/concepts/collections/#setting-up-multitenancy
        self.collection_index = len(self.client.get_collections().collections)

        # Note: Qdrant has an internal definition of "shard" that is different than
        # what is currently used by Hyrax (specifically ChromaDB). Here we set
        # shard_number to CPU core count * 2. Qdrant docs advocate starting with 12:
        # https://qdrant.tech/documentation/guides/distributed_deployment/#choosing-the-right-number-of-shards
        self.collection_name = f"shard_{self.collection_index}"
        created_collection = None
        if not self.client.collection_exists(self.collection_name):
            created_collection = self.client.create_collection(
                collection_name=self.collection_name,
                vectors_config=models.VectorParams(
                    #! This stinks - we should just check the size of the data
                    #! when we call `save_to_database` and then set this automatically
                    #! as a parameter in self.context["blah"] or something.
                    size=self.config["vector_db.qdrant"]["vector_size"],
                    distance=models.Distance.EUCLID,
                    on_disk=True,
                ),
                shard_number=os.cpu_count() * 2,
            )

        if not created_collection:
            raise RuntimeError(f"Failed to create collection {self.collection_name} in Qdrant.")

        self.collection_size = self.client.count(collection_name=self.collection_name, exact=True)

        return self.collection_name

    def insert(self, ids: list[Union[str, int]], vectors: list[np.ndarray]):
        """Insert a batch of vectors into the Qdrant database.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids to associate with the vectors
        vectors : list[np.ndarray]
            The vectors to insert into the database
        """
        if self.client is None:
            self.connect()

        expected_size = self.config["vector_db.qdrant"]["vector_size"]
        for idx, vector in enumerate(vectors):
            if len(vector) != expected_size:
                raise ValueError(
                    f"Vector at index {idx} has size {len(vector)}, but expected size is {expected_size}."
                )

        uuids = [self._convert_id_to_uuid(i) for i in ids]
        # Insert data into the collection
        self.client.upsert(
            collection_name=self.collection_name,
            points=models.Batch(
                ids=uuids,
                vectors=vectors,
                payloads=[{"id": id} for id in ids],
            ),
        )

        # Update the collection size after insertion
        self.collection_size = self.client.count(collection_name=self.collection_name, exact=True).count
        return self.collection_size

    def search_by_id(self, id: Union[str, int], k: int = 1) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given id in the database.

        Qdrant will exclude the id itself from the results, thus we first
        retrieve the vector for a given id, and then use that vector to find the
        k nearest neighbors.

        Parameters
        ----------
        id : Union[str, int]
            The id of the vector in the database for which we want to find the
            k nearest neighbors
        k : int, optional
            The number of nearest neighbors to return, by default 1, return only
            the closest neighbor

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input vector id as the key and the ids of the k
            nearest neighbors as the value.
        """
        if self.client is None:
            self.connect()

        # Retrieve the vector for the given id
        query_vector = self.get_by_id(id)

        # Find the k nearest neighbors for that vector
        res = {id: self._query_by_vector(query_vector[id], k)}
        return res

    def search_by_vector(
        self, vectors: Union[np.ndarray, list[np.ndarray]], k: int = 1
    ) -> dict[int, list[Union[str, int]]]:
        """Get the ids of the k nearest neighbors for a given vector.

        Parameters
        ----------
        vectors : Union[np.array, list[np.ndarray]]
            The one or more vectors to use when searching for nearest neighbors
        k : int, optional
            The number of nearest neighbors to return, by default 1, return only
            the closest neighbor

        Returns
        -------
        dict[int, list[Union[str, int]]]
            Dictionary with input vector index as the key and the ids of the
            k nearest neighbors as the value.
        """
        if self.client is None:
            self.connect()

        # If a single vector is provided, convert it to a list
        if isinstance(vectors, np.ndarray):
            vectors = [vectors]

        # Find the k nearest neighbors for the provided vector
        res = {i: self._query_by_vector(v, k) for i, v in enumerate(vectors)}
        return res

    def _query_by_vector(self, vector: np.ndarray, k: int = 1) -> list[str]:
        """Query the Qdrant database for the k nearest neighbors of a given vector."""

        query_results = self.client.query_points(
            collection_name=self.collection_name,
            query=models.NearestQuery(nearest=vector),
            search_params=models.SearchParams(),
            limit=k,
        )

        return [point.payload["id"] for point in query_results.points]

    def get_by_id(self, ids: list[Union[str, int]]) -> dict[Union[str, int], list[float]]:
        """Retrieve the vectors associated with a list of ids.

        Parameters
        ----------
        ids : list[Union[str, int]]
            The ids of the vectors to retrieve.

        Returns
        -------
        dict[Union[str, int], list[float]]
            Dictionary with the ids as the keys and the vectors as the values.
        """
        if self.client is None:
            self.connect()

        if not isinstance(ids, list):
            ids = [ids]

        uuids = [self._convert_id_to_uuid(i) for i in ids]

        points = self.client.retrieve(
            collection_name=self.collection_name,
            ids=uuids,
            with_vectors=True,
            with_payload=True,
        )

        # Return the vectors for the given ids
        return {point.payload["id"]: point.vector for point in points}



import logging
import os

import pytest

from hyrax.config_utils import ConfigDict, ConfigManager


def test_merge_configs():
    """Basic test to ensure that the merge_configs function will join two dictionaries
    correctly, meaning:
    1) The user_config values should override the default_config values.
    2) Values in the default_config that are not in the user_config should remain unchanged.
    3) Values in the user_config that are not in the default_config should be added.
    4) Nested dictionaries should be merged recursively.
    """
    default_config = {
        "a": 1,
        "b": 2,  # This tests case 2
        "c": {"d": 3, "e": 4},
    }

    user_config = {
        "a": 5,  # This tests case 1
        "c": {
            "d": 6  # This tests case 4
        },
        "f": 7,  # This tests case 3
    }

    expected = {"a": 5, "b": 2, "c": {"d": 6, "e": 4}, "f": 7}

    assert ConfigManager.merge_configs(default_config, user_config) == expected


def test_get_runtime_config():
    """Test that the get_runtime_config function will load the default and user defined
    runtime configuration files, merge them, and return the final configuration as a
    dictionary.
    """

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    expected = {
        "general": {"dev_mode": True},
        "train": {
            "model_name": "example_model",
            "model_class": "new_thing.cool_model.CoolModel",
            "model": {"weights_filepath": "final_best.pth", "layers": 3},
        },
        "infer": {"batch_size": 8},
        "bespoke_table": {"key1": "value1", "key2": "value2"},
    }

    string_representation = """# this is the default config file
[general]
# set dev_mode to true when developing
# set to false for production use
dev_mode = true

[train]
model_name = "example_model" # Use a built-in Hyrax model
model_class = "new_thing.cool_model.CoolModel" # Use a custom model

[train.model]
weights_filepath = "final_best.pth"
layers = 3


[infer]
batch_size = 8 # change batch size

[bespoke_table]
# this is a bespoke table
key1 = "value1"
key2 = "value2" # unlikely to modify
"""
    assert runtime_config == expected
    assert runtime_config.as_string() == string_representation


def test_validate_runtime_config():
    """Test that the validate_runtime_config function will raise a RuntimeError
    if a user key is not defined in the default configuration dictionary.
    """

    default = {"general": {"dev_mode": False}, "train": {"model_name": "example_model"}}
    default_config = ConfigDict(default)

    user = {"general": {"dev_mode": False, "foo": "bar"}}
    user_config = ConfigDict(user)

    with pytest.raises(RuntimeError) as excinfo:
        ConfigManager._validate_runtime_config(user_config, default_config)

    assert "Runtime config contains key" in str(excinfo.value)


def test_validate_runtime_config_section():
    """Test that the validate_runtime_config function will raise a RuntimeError
    if a user section name conflicts with a default configuration key.
    """

    default = {"general": {"dev_mode": False}, "train": {"model_name": "example_model"}}
    default_config = ConfigDict(default)

    user = {"general": {"dev_mode": {"b": 2}}}
    user_config = ConfigDict(user)

    with pytest.raises(RuntimeError) as excinfo:
        ConfigManager._validate_runtime_config(user_config, default_config)

    assert "Runtime config contains a section named dev_mode" in str(excinfo.value)


def test_find_external_library_config_path_no_module():
    """Test that a ModuleNotFound error is raised when trying to import a
    non-existent module.
    """

    config = {"general": {"dev_mode": False}, "model": {"name": "foo.bar.model.Model"}}
    default_config = ConfigDict(config)

    with pytest.raises(ModuleNotFoundError) as excinfo:
        ConfigManager._find_external_library_default_config_paths(default_config)

    assert "Check installation" in str(excinfo.value)


def test_find_external_library_config_path_no_default_config(caplog):
    """Test that a warning is logged when a default configuration file is not found."""

    config = {"general": {"dev_mode": False}, "model": {"name": "toml.bar.model.Model"}}
    default_config = ConfigDict(config)

    with caplog.at_level(logging.WARNING):
        ConfigManager._find_external_library_default_config_paths(default_config)

    assert "default_config.toml" in caplog.text


def test_config_help(capsys):
    """Basic use case where config help function prints entire config"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help()

    captured = capsys.readouterr()

    expected_output = """# this is the default config file
[general]
# set dev_mode to true when developing
# set to false for production use
dev_mode = true

[train]
model_name = "example_model" # Use a built-in Hyrax model
model_class = "new_thing.cool_model.CoolModel" # Use a custom model

[train.model]
weights_filepath = "final_best.pth"
layers = 3


[infer]
batch_size = 8 # change batch size

[bespoke_table]
# this is a bespoke table
key1 = "value1"
key2 = "value2" # unlikely to modify
"""

    assert expected_output in captured.out


def test_config_help_specific_table(capsys):
    """Basic use case where config help function prints one table"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help("bespoke_table")

    captured = capsys.readouterr()

    expected_output = """[bespoke_table]
# this is a bespoke table
key1 = "value1"
key2 = "value2" # unlikely to modify
"""

    assert expected_output in captured.out


def test_config_help_table_and_key(capsys):
    """Basic use case where config help function prints table when provided with
    a table and key."""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help("train", "model_name")

    captured = capsys.readouterr()

    expected_output = """[train]
model_name = "example_model" # Use a built-in Hyrax model
model_class = "new_thing.cool_model.CoolModel" # Use a custom model
"""

    assert expected_output in captured.out


def test_config_help_non_existant_table_or_key(capsys):
    """Basic use case where config help function prints error message when table
    or key isn't present"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    non_table_name = "non_existant_table"
    runtime_config.help(non_table_name)

    captured = capsys.readouterr()

    expected_output = f"Could not find '{non_table_name}' in the config"

    assert expected_output in captured.out


def test_config_help_real_table_non_existant_key(capsys):
    """User requests help for real table, but non-existant key"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help("general", "non_existant_key")

    captured = capsys.readouterr()

    expected_output = "Cannot find ['general']['non_existant_key'] in the current configuration."

    assert expected_output in captured.out


def test_config_help_key_in_multiple_tables(capsys):
    """User requests help for a key that is present in more than one table"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config_repeated_keys.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help("name")

    captured = capsys.readouterr()

    expected_output = """Found 'name' in the following config sections: [model], [loss], [optimizer]
[model]
name = "resnet"
layers = 3


[loss]
name = "cross_entropy"


[optimizer]
name = "adam"
"""

    assert expected_output in captured.out


def test_config_help_too_many_args(capsys):
    """User requests help for with >2 input args"""

    this_file_dir = os.path.dirname(os.path.abspath(__file__))
    config_manager = ConfigManager(
        runtime_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_user_config.toml")
        ),
        default_config_filepath=os.path.abspath(
            os.path.join(this_file_dir, "./test_data/test_default_config.toml")
        ),
    )

    runtime_config = config_manager.config

    runtime_config.help("general", "dev_mode", "extra_arg")

    captured = capsys.readouterr()

    expected_output = """Too many arguments provided. Expecting 0, 1, or 2 arguments.
Usage: config.help(['table_name'|'key_name']), config.help('table_name', 'key_name')"""

    assert expected_output in captured.out



# ruff: noqa: D101, D102
import logging

import numpy as np
from torch.utils.data import Dataset, IterableDataset

from hyrax.config_utils import ConfigDict

from .data_set_registry import HyraxDataset

logger = logging.getLogger(__name__)


class HyraxCifarBase:
    """Base class for Hyrax Cifar datasets"""

    def __init__(self, config: ConfigDict):
        import torchvision.transforms as transforms
        from astropy.table import Table
        from torchvision.datasets import CIFAR10

        transform = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
        )
        self.cifar = CIFAR10(
            root=config["general"]["data_dir"], train=True, download=True, transform=transform
        )
        metadata_table = Table(
            {"label": np.array([self.cifar[index][1] for index in range(len(self.cifar))])}
        )
        super().__init__(config, metadata_table)


class HyraxCifarDataSet(HyraxCifarBase, HyraxDataset, Dataset):
    """Map style CIFAR 10 dataset for Hyrax

    This is simply a version of CIFAR10 that is initialized using Hyrax config with a transformation
    that works well for example code.

    We only use the training split in the data, because it is larger (50k images). Hyrax will then divide that
    into Train/test/Validate according to configuration.
    """

    def __len__(self):
        return len(self.cifar)

    def __getitem__(self, idx):
        image, label = self.cifar[idx]
        return {
            "object_id": idx,
            "image": image,
            "label": label,
        }


class HyraxCifarIterableDataSet(HyraxCifarBase, HyraxDataset, IterableDataset):
    """Iterable style CIFAR 10 dataset for Hyrax

    This is simply a version of CIFAR10 that is initialized using Hyrax config with a transformation
    that works well for example code. This version only supports iteration, and not map-style access

    We only use the training split in the data, because it is larger (50k images). Hyrax will then divide that
    into Train/test/Validate according to configuration.
    """

    def __iter__(self):
        for idx, (image, label) in enumerate(self.cifar):
            yield {
                "object_id": idx,
                "image": image,
                "label": label,
            }
